//**************************************************************
//*
//* Pinbelegung
//*    ----------------------------------
//*        1  VDD (+5V)
//*        2  GP5 -out- (230V Last Relais)
//*        3  GP4 -out- (LED - rot)
//*        4  GP3 -in - (lernen)
//*        5  AN2 -in - (Messspannung)
//*        6  GP1 -out- (LED - grn)
//*        7  GP0 -in - (zero-crossing opto)
//*        8  VSS (GND)
//*
//**************************************************************
//
// loetmeister.de - 2014, v0.5
//
// AUS-Schalter 230V - 0Watt (Mottai multi - Mod)
//
// PIC 12F675
// Strommessung von 230V Verbrauchern (z.B. PC) ber einen kleinen 'Trafo'
// (Stromwander - current transformer (CT)) und ADC.
// Wird die im EEPROM abgespeicherte Schwelle unterschritten, wird der
// Verbraucher und die Schaltung selbst vom Strom getrennt.
//
// http://www.loetmeister.de/Elektronik/ausschalter_230v_mottai_multi_mod.htm
//
// Prozessortakt:  4 MHz intern
// Eingangsspannung: 5V

/******************************************************************************/
/* Files to Include                                                           */
/******************************************************************************/

#include <xc.h>         /* XC8 General Include File */

#include <stdint.h>        /* For uint8_t definition */
#include <stdbool.h>       /* For true/false definition */


#include "system.h"        /* System funct/params, like osc/peripheral config */
#include "user.h"          /* User funct/params, such as InitApp */

/******************************************************************************/
/* User Global Variable Declaration                                           */
/******************************************************************************/
unsigned int ADCresult_buf[const_ADCsamples]=0; // array for ADC result (moving average)
unsigned int ADCresult_sum=0, ADCresult;

char ADCindex=0;
char timer2=const_timer2;
char learnButton_timer2=const_learnButton_timer2;
char sleepCounter=const_sleepCounter;

bit learnButtonState;
bit learnButtonStateOld;
bit LEDflash;
bit programmingDone=0;
bit relaisOperation;
bit zeroCrossingState;


// to store the ADC result values
union {unsigned int ADRES;
    unsigned char ad_hi_low[2];
} ADCout, powerOffLevel;

/******************************************************************************/
/* Main Program                                                               */
/******************************************************************************/
void main(void)
{
    /* Configure the oscillator for the device */
    //ConfigureOscillator(); // no need here, default startup code will do it

    /* Initialize I/O and Peripherals */
    InitApp();

    
    outLEDredPin = 1; // turn LED on to show powerup

    // wait about 4 seconds (no poweron by accident)
    for (ADCresult_sum = 31000; ADCresult_sum!=0; --ADCresult_sum)
    {
        _delay(127); // 127s (+ about 60ms for the "for" loop execution)
    }
    ADCresult_sum = 0; // we "abused" this variable, so set it to a proper value for the main loop

#ifndef ENABLE_AC_ZERO_DEDECTION
    /* set output pin HIGH to turn on the mains relais */
    outRelaisPin = RelaisOn;
#else
    /* switch mains relay at AC zero crossing */
    // TODO ?digital in: enable interrupt, wait for it (means AC peak - or close to it), diable interrupt, turn on relais with offset to swith at AC zero
    // analog in: read ADC, wait for raising voltage - then for the drop (should be close after the AC peak), turn on relais with offset to swith at AC zero
    // read ADC, detect ~50Hz and switch relay - considering the relay switching time

    relaisOperation = RelaisOn;
    activate_ac_zero_interrupt();    // switch relais in ISR, but only return if interrupt was disabled again (i.e. operation complete)

#endif

    /* read EEPROM values */
    powerOffLevel.ad_hi_low[0] = eeprom_read(0x0); // read L part
    powerOffLevel.ad_hi_low[1] = eeprom_read(0x1); // read H part

    // all init done, allow interrupts
    ei(); // enable all interrupts (actually not needed, the EEPROM read routine sets GIE=1 when done)

    // main loop
    while(1)
    {
        /* turn off when sleep delay is over */
        if (sleepCounter <= 0)
        {
#ifndef ENABLE_AC_ZERO_DEDECTION
            outRelaisPin = RelaisOff; // turn mains off
#else
            relaisOperation = RelaisOff;
            activate_ac_zero_interrupt();    // switch relais in ISR, but only return if interrupt was disabled again (i.e. operation complete)
#endif
        }
        
        // sample input for analog channel 2 (Transformer voltage ~ power consumption)
        // DC voltage, max. 3.7V @ 60W
        ADCON0bits.CHS = 0b10;    // select Analog Channel AN2

        _delay(16); // wait 16 cycles (1/(4MHz/4) = 1s * 16 = 16 s
        adc_read();
        ADCout.ad_hi_low[0] = ADRESL;
        ADCout.ad_hi_low[1] = ADRESH;

        
        /* button read section */
        learnButtonState = inLearnButtonPin;  // read input
        
        // check if start button is still pressed
        if (learnButtonState != learnButtonStateOld)
        {
            learnButton_timer2 = const_learnButton_timer2; // reset learnButton_timer2 if not
            programmingDone = 0;
        }
        /* learnButton_timer2 decrement by one at every timer2 role over (4,1Hz) delay is defnied by const_learnButton_timer2  */
        if (learnButton_timer2 <= 0 && inLearnButtonPin == LearnButtonPressed && !programmingDone)  // we only want to know when button is pressed
        {
//            powerOffLevel.ADRES = ADCout.ADRES;  // use current ADC value as new poweroff value
            powerOffLevel.ADRES = ADCresult;  // use new poweroff value
            
            // when learn button was pressed and hold, store the current ADC reading in the EEPROM
            // PIC12F629/675  devices  have  128 bytes of data EEPROM with an address range from 0h to 7Fh
            eeprom_write(0x0, powerOffLevel.ad_hi_low[0]);
            eeprom_write(0x1, powerOffLevel.ad_hi_low[1]);

            programmingDone = 1;  // just saved the poweroff level. Don't do it again, until the learn button was released
        }
        learnButtonStateOld = learnButtonState;

    } // while
}

/******************************************************************************/
/* Interrupt routine                                                          */
/* Using T0IF (timer0 interrupt), GPIF (GPIO port change)                     */
/******************************************************************************/
void interrupt isr(void)
{
#ifdef ENABLE_AC_ZERO_DEDECTION
    if (INTCONbits.GPIF)    // this interrupt is only enabled when the relais need to be operated
    {
        if (relaisOperation == RelaisOn)
        {
            _delay(RelaisOn_delay_mikrosec);    // wait, to operate when AC = 0
            outRelaisPin = RelaisOn;    // Operate time (set time) depdens on relais type
        }
        else
        {
            _delay(RelaisOff_delay_mikrosec);    // wait, to operate when AC = 0
            outRelaisPin = RelaisOff;   // Release time depends on relais type
        }

        INTCONbits.GPIE = 0;    // disable GPIO port change interrupt
        INTCONbits.T0IE = 1;    // Timer0 interupt erlauben
        INTCONbits.GPIF = 0;    // clear interrupt flag
    } // GPIF
#endif

    // reset timer0 when interrupt has occured (62,5Hz)
    if (INTCONbits.T0IF)
    {
        TMR0 = const_timer0;    // reset
        --timer2;

        /* calculate the ADC average of the last n results */
        ADCresult_sum += (ADCout.ADRES - ADCresult_buf[ADCindex]);  // add new and substract old value
        ADCresult_buf[ADCindex] = ADCout.ADRES;  // update buffer with current reading
        ADCresult = ADCresult_sum / const_ADCsamples;  // calculate average

        ++ADCindex;
        if (ADCindex > const_ADCsamples)
        {
            ADCindex = 0; // reset when last array element was processed
        }

        // check if timer2 ticked down (4,1667Hz)
        if (timer2 == 0)
        {
            timer2 = const_timer2;  // reset timer2

            if (learnButtonState == LearnButtonPressed && !programmingDone)
            {
                outLEDredPin = 0;
                --learnButton_timer2;
                outLEDgreenPin = !outLEDgreenPin; // flash LED fast to indicate coming learning
            }
            else    // if not in learn mode: check if ADC result is below poweroff threshold
            {
                // tick down counter to delay mains shutdown
                if ( ADCresult <= (powerOffLevel.ADRES + 6) )
                {
                    --sleepCounter;
                    outLEDgreenPin = 0;

                    /* change LED status every second time to blink slowly */
                    if (LEDflash == 1)
                    {
                        outLEDredPin = !outLEDredPin; // flash LED to indicate coming shutdown
                    }
                    else
                    {
                        LEDflash = !LEDflash;
                    }
                }
                else
                {
                    outLEDredPin = 0;
                    sleepCounter = const_sleepCounter;  // reset counter
                    outLEDgreenPin = 1; // keep LED on to show normal operation
                }
            }
        }
    } //T0IF
    INTCONbits.T0IF = 0;   // clear interrupt flag
}


/******************************************************************************/
/* Enables the interrupt for the AC zero crossing detection                   */
/* Using port change interrupt GPIF/IOC                                       */
/* only return if interrupt was disabled again */
/******************************************************************************/
#ifdef ENABLE_AC_ZERO_DEDECTION
void activate_ac_zero_interrupt(void)
{
    // turn off other interrupts?
    di();   //disable all interrupts
    INTCONbits.T0IE = 0;    // Timer0 interupt deaktivieren

    // wait until input pin becommes low, so the next pin change interrupt will occur on rising edge
    while (zeroCrossingState == 1)
    {
        zeroCrossingState = in0crossingPin;
    }

    INTCONbits.GPIF = 0;    // clear interrupt flag
//    IOCbits.IOC0 = 1;   // allow interrupt on change @GP0
    INTCONbits.GPIE = 1;    // eable GPIO port change interrupt
    ei(); // enable all interrupts (actually not all, as Timer0 is disabled and will only be enabled again after port interrupt was handelt completely)
    
    while(INTCONbits.GPIE)  // sit and wait until port change interrupt is disabled again
        ;
}
#endif

/******************************************************************************/
// EOF